/* p80211.c
   802.11 header/frame handling

   Author : KoreK
   Date : Sep 1, 2004
   Licence : GPL v2 or above (see accompanying COPYING file)

   Copyright (c) 2004 KoreK, All Rights reserved
*/

#define P80211_C

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>

#include "p80211.h"
#include "p80211_const.h"

/***********/
/* setters */
void p80211_header_init(struct p80211_header *h)
{
    memset(h,0,sizeof *h);
}


void p80211_set_type(struct p80211_header *h,
		     int type)
{
    if (type != INVALID_FRAME) {
	h->fc=(h->fc & fc_type_mask) | fc_type[type];
    }
}


void p80211_set_flag(struct p80211_header *h,
		     int flag,
		     int value)
{
    if (value) h->fc=h->fc | flag;
    else       h->fc=h->fc & ~flag;
}


void p80211_set_duration(struct p80211_header *h,
			 int duration)
{
    h->dur_id=h_to_le(duration);
}


void p80211_set_fragnb(struct p80211_header *h,
		       int frag_nb)
{
    int seq;
    seq = le_to_h(h->seq);
    seq = (seq & 0xfff0) | (frag_nb & 0x000f);
    h->seq = h_to_le(seq);
}


void p80211_set_seqnb(struct p80211_header *h,
		      int seqnb)
{
    int seq;
    seq = le_to_h(h->seq);
    seq = (seq & 0x000f) | ((seqnb << 4) & 0xfff0);
    h->seq = h_to_le(seq);
}


/* given a valid header, fillup mac_idx=DA,SA,BSSID,... with MAC
   return 0 if successful */
int p80211_set_mac(struct p80211_header *h,
		   int mac_idx,
		   const u_char *MAC)
{
    u_int offset;
    struct p80211_frame_struct fs;

    fs = p80211_struct(h);
    if (fs.type == INVALID_FRAME) return -1;

    offset=mac_offset[fs.mac_order][mac_idx];
    if (offset == 0) return -1;

    memcpy(h->base+offset,MAC,6);
    return 0;
}




/***********/
/* getters */
struct p80211_frame_struct p80211_struct(const struct p80211_header *h)
{
    struct p80211_frame_struct fs;
    u_int fc;

    fc = le_to_h(h->fc);
    fs = type_DS[(fc>>2) & 0x3f][(fc>>8) & 0x03];
    return fs;
}


int p80211_type(const struct p80211_header *h)
{
    struct p80211_frame_struct fs;
    int fc;

    fc = le_to_h(h->fc);
    fs = type_DS[(fc>>2) & 0x3f][(fc>>8) & 0x03];
    return fs.type;
}


int p80211_flag(const struct p80211_header *h,
		int flag)
{
    return h->fc & flag;
}


int p80211_duration(const struct p80211_header *h)
{
    return le_to_h(h->dur_id);
}


int p80211_fragnb(const struct p80211_header *h)
{
    return le_to_h(h->seq) & 0x000f;
}


int p80211_seqnb(const struct p80211_header *h)
{
    return (le_to_h(h->seq) >> 4) & 0x0fff;
}


const u_char *p80211_mac(const struct p80211_header *h,
			 int mac_idx)
{
    struct p80211_frame_struct fs;
    u_int offset;
 
    fs = p80211_struct(h);
    if (fs.type == INVALID_FRAME) return 0;

    offset=mac_offset[fs.mac_order][mac_idx];
    if (offset == 0) return 0;
    else return h->base+offset;
}


int p80211_data_len(const struct p80211_header *h)
{
    return le_to_h(h->data_len);
}




/******************************/
/* frame creation/destruction */
p80211_frame_t make_frame(struct p80211_header *h,
				const u_char *data,
				size_t data_len)
{
    p80211_frame_t f;

    f=(p80211_frame_t )malloc(sizeof(*h) + data_len);
    if (f==0) return 0;

    if (h) memcpy(f,h,sizeof *h);
    else p80211_header_init(f);

    f->data_len=h_to_le(data_len);
    if (data_len>0) {
	if (data) memcpy(f->data,data,data_len);
	else memset(f->data,0,data_len);
    }
    return f;
}

void free_frame(p80211_frame_t f)
{
    free(f);
}

// int frame_valid(p80211_frame_t f)




/* dumpers */
/* print some char buffer, the standard way, ie 16 char / line, white-spaced
   with double space in the middle */
void fprint_hex_buffer(FILE *f,
		       const u_char *b,
		       size_t b_len)
{
    size_t i;
    for (i=0; i<b_len;i++) {
	if (i != 0) {
	    switch (i % 16) {
		case 0:
		    fprintf(f,"\n");
		    break;
		case 8:
		    fprintf(f,"  ");
		    break;
		default:
		    fprintf(f," ");
	    }
	}
	fprintf(f,"%02hhx",b[i]);
    }
    fprintf(f,"\n");
}


void fprint_header(FILE *f,
		   const struct p80211_header *h)
{
    int fc;
    struct p80211_frame_struct fs;
    int i;
    size_t mo;

    fc=le_to_h(h->fc);
    fs=type_DS[(fc>>2) & 0x3f][(fc>>8) & 0x03];
    /* header */
    fprintf(f,"Frame type: %-26s data_len: %-18d ver: %d\n",
	    type_name[fs.type],
	    le_to_h(h->data_len),
	    fc & 0x03);
    /* flags */
    fprintf(f,"toDS:%d  fromDS:%d  morefrag:%d  retry:%d  pwrmgt:%d  "
	    "moredata:%d  wep:%d  order:%d\n",
	    p80211_flag(h,toDS_f)!=0,p80211_flag(h,fromDS_f)!=0,
	    p80211_flag(h,morefrag_f)!=0,p80211_flag(h,retry_f)!=0,
	    p80211_flag(h,pwrmgt_f)!=0,p80211_flag(h,moredata_f)!=0,
	    p80211_flag(h,wep_f)!=0,p80211_flag(h,order_f)!=0);

    if (fs.type == INVALID_FRAME) {
	fprint_hex_buffer(f,(u_char *)h,sizeof *h);
    }
    else {
	/* print out macs */
	for (i=0; i<5; i++) {
	    mo=mac_offset[fs.mac_order][i];
	    if (mo) {
		const char *mac;
		mac = h->base + mo;
		fprintf(f,
			"%-3s: %02hhx%02hhx%02hhx%02hhx%02hhx%02hhx  ",
			mac_name[i],
			mac[0],mac[1],mac[2],mac[3],mac[4],mac[5]);
	    }
	}
	fprintf(f,"\n");

	/* print duration / frag # / seq # */
	fprintf(f,"duration: %5d   frag#: %4d   seq#: %4d\n",
		p80211_duration(h),p80211_fragnb(h),p80211_seqnb(h));
    }
}


/* avoid calling on an unconstructed frame */
void fprint_frame(FILE *f,
		  const p80211_frame_t frame)
{
    fprint_header(f,frame);
    fprint_hex_buffer(f,frame->data,frame->data_len);
}

